import os
import numpy as np

from astropy.io import fits
from astropy.table import Table
from astropy import wcs
from astropy.coordinates import SkyCoord

from .utils_for_test import create_test_ffis
from ..make_cube import CubeFactory
from ..cube_cut import CutoutFactory

def checkcutout(cutfile,pixcrd,world,csize,ecube,eps=1.e-7):
    """Check FITS cutout for correctness
    
    Checks RA_OBJ/DEC_OBJ in primary header, and TIME, FLUX, and
    FLUX_ERR in table.
    
    Inputs:
    cutfile  Name of FITS cutout file
    pixcrd   [2] pixel coordinates for cutout center [cy,cx]
    world    [2] RA, Dec in degrees for cutout center
    csize    Integer size of cutout (probably should be odd)
    ecube    Simulated data cube
    eps      Maximum allowed distance offset in degrees
    Returns True on success, False on failure
    """
    
    ix = int(pixcrd[1])
    iy = int(pixcrd[0])
    x1 = ix - csize//2
    x2 = x1 + csize - 1
    y1 = iy - csize//2
    y2 = y1 + csize - 1
    hdulist = fits.open(cutfile)
    ra_obj = hdulist[0].header['RA_OBJ']
    dec_obj = hdulist[0].header['DEC_OBJ']
    pinput = SkyCoord(world[0],world[1],frame='icrs',unit='deg')
    poutput = SkyCoord(ra_obj,dec_obj,frame='icrs',unit='deg')
    
    dist = pinput.separation(poutput).degree
    assert dist <= eps, "{} separation in primary header {} too large".format(cutfile,dist)
        
    cx = ecube.shape[0]
    cy = ecube.shape[1]
    ntimes = ecube.shape[2]
    tab = hdulist[1].data
    assert len(tab) == ntimes, "{} expected {} entries, found {}".format(cutfile,ntimes,len(tab))
    assert (tab['TIME']==(np.arange(ntimes)+0.5)).all(), "{} some time values are incorrect".format(cutfile)

    check1(tab['FLUX'],x1,x2,y1,y2,ecube[:,:,:,0],'FLUX',cutfile)
    check1(tab['FLUX_ERR'],x1,x2,y1,y2,ecube[:,:,:,1],'FLUX_ERR',cutfile)
    
    return 

def check1(flux,x1,x2,y1,y2,ecube,label,cutfile):
    """Test one of flux or error"""
    cx = ecube.shape[0]
    cy = ecube.shape[1]
    if x1 < 0:
        assert np.isnan(flux[:,:-x1,:]).all(), "{} {} x1 NaN failure".format(cutfile,label)
    
    if y1 < 0:
        assert np.isnan(flux[:,:,:-y1]).all(), "{} {} y1 NaN failure".format(cutfile,label)
        
    if x2>=cx:
        assert np.isnan(flux[:,-(x2-cx+1):,:]).all(), "{} {} x2 NaN failure".format(cutfile,label)
        
    if y2>=cy:
        assert np.isnan(flux[:,:,-(y2-cy+1):]).all(), "{} {} y2 NaN failure".format(cutfile,label)
        
    x1c = max(x1,0)
    y1c = max(y1,0)
    x2c = min(x2,cx-1)
    y2c = min(y2,cy-1)
    scube = ecube[x1c:x2c,y1c:y2c,:]
    sflux = np.moveaxis(flux[:,x1c-x1:x2c-x1,y1c-y1:y2c-y1],0,-1)
    assert (scube==sflux).all(), "{} {} comparison failure".format(cutfile,label)

    return



def test_cube_cutout(tmpdir):
    """
    Testing the cube cutout functionality.
    """

    # Making the test cube
    cube_maker = CubeFactory()
    
    img_sz = 10
    num_im = 100
    
    ffi_files = create_test_ffis(img_sz, num_im)
    cube_file = cube_maker.make_cube(ffi_files, "make_cube-test-cube", verbose=False)

    # Read one of the input images to get the WCS
    img_header = fits.getheader(ffi_files[0], 1)
    cube_wcs = wcs.WCS(img_header)

    # get pixel positions at edges and center of image
    # somewhat cryptic one-liner to get the grid of points
    pval = np.array([0,img_sz//2,img_sz-1],dtype=np.float)
    pixcrd = pval[np.transpose(np.reshape(np.mgrid[0:3,0:3],(2,9)))]
    
    # add one more giant cutout that goes off all 4 edges
    pixcrd = np.append(pixcrd,pixcrd[4].reshape(1,2),axis=0)

    # getting the world coordinates
    world_coords = cube_wcs.all_pix2world(pixcrd,0)

    # Getting the cutouts
    cutbasename = 'make_cube-cutout{}.fits'
    cutlist = [cutbasename.format(i) for i in range(len(world_coords))]
    csize = [img_sz//2]*len(world_coords)
    csize[-1] = img_sz+5
    for i, v in enumerate(world_coords):
        coord = SkyCoord(v[0],v[1],frame='icrs',unit='deg')
        cut = CutoutFactory().cube_cut(cube_file,coord,csize[i],target_pixel_file=cutlist[i],verbose=False)

    # expected values for cube
    ecube = np.zeros((img_sz,img_sz,num_im,2))
    plane = np.arange(img_sz*img_sz,dtype=np.float32).reshape((img_sz,img_sz))
    for i in range(num_im):
        ecube[:,:,i,0] = -plane
        ecube[:,:,i,1] = plane
        plane += img_sz*img_sz

    # Doing the actual checking
    for i, cutfile in enumerate(cutlist):
        checkcutout(cutfile,pixcrd[i],world_coords[i],csize[i],ecube)
           
    



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""This module implements the cutout functionality."""

import numpy as np
import astropy.units as u

from astropy.io import fits
from astropy.table import Table,Column
from astropy.coordinates import SkyCoord
from astropy import wcs

from time import time
from copy import deepcopy

import os
import warnings

from .exceptions import InputWarning, TypeWarning, InvalidQueryError



class CutoutFactory():
    """
    Class for creating image cutouts.

    This class emcompasses all of the cutout functionality.  
    In the current version this means creating cutout target pixel files from TESS full frame image cubes.
    Future versions will include more generalized cutout functionality.
    """

    def __init__(self):
        """
        Initiazation function.
        """

        self.cube_wcs = None # WCS information from the image cube
        self.cutout_lims = np.zeros((2,2),dtype=int) #  Cutout pixel limits, [[ymin,ymax],[xmin,xmax]]
        self.center_coord = None # Central skycoord

        

    
    def _get_cube_wcs(self, table_header,table_row):
        """
        Takes the header and one entry from the cube table of image header data and  builds 
        a WCS object that encalpsulates the given WCS information.  

        This object is stored in ``self.cube_wcs``.

        Parameters
        ----------
        table_header : `~astropy.io.fits.Header`
            The cube image header data table header.
        table_row : `~astropy.io.fits.FITS_record`
            One row from the cube image header data table.

        Returns
        -------
        response : `~astropy.wcs.WCS`
            WCS object containing the WCS information from the given cube image header data table row.
        """
    
        wcs_header = fits.header.Header()

        for header_key, header_val in table_header.items():
            if not 'TTYPE' in header_key:
                continue
        
            col_num = int(header_key[5:]) - 1
            tform = table_header[header_key.replace("TTYPE","TFORM")]
            if "A" in tform:
                wcs_header[header_val] =  str(table_row[col_num])
            elif "D" in tform:
                wcs_header[header_val] =  float(table_row[col_num])
            elif "J" in tform:
                wcs_header[header_val] =  int(table_row[col_num])
            else:
                warnings.warn("Unknown data type, keyword value will be parsed as a string.",
                              TypeWarning)
           
        self.cube_wcs =  wcs.WCS(wcs_header)


    def _get_cutout_limits(self, cutout_size):
        """
        Takes the center coordinates, cutout size, and the wcs from
        which the cutout is being taken and returns the x and y pixel limits
        for the cutout.

        Parameters
        ----------
        center_coord : `~astropy.coordinates.SkyCoord`
            The central coordinate for the cutout
        cutout_size : array
            [ny,nx] in with ints (pixels) or astropy quantities

        Returns
        -------
        response : `numpy.array`
            The cutout pixel limits in an array of the form [[ymin,ymax],[xmin,xmax]]
        """
        
        # Note: This is returning the center pixel in 1-up
        center_pixel = self.center_coord.to_pixel(self.cube_wcs, 1)

        lims = np.zeros((2,2),dtype=int)

        for axis, size in enumerate(cutout_size):
        
            if not isinstance(size, u.Quantity): # assume pixels
                dim = size/2
            elif size.unit == u.pixel: # also pixels
                dim = size.value/2
            elif size.unit.physical_type == 'angle':
                pixel_scale = u.Quantity(wcs.utils.proj_plane_pixel_scales(self.cube_wcs)[axis], self.cube_wcs.wcs.cunit[axis])
                dim = (size / pixel_scale).decompose()/2

            lims[axis,0] = int(np.round(center_pixel[axis] - 1 - dim))
            lims[axis,1] = int(np.round(center_pixel[axis] - 1 + dim))

        # Checking at least some of the cutout is on the cube
        if ((lims[0,0] <= 0) and (lims[0,1] <=0)) or ((lims[1,0] <= 0) and (lims[1,1] <=0)):
            raise InvalidQueryError("Cutout location is not in cube footprint!") 
    
        self.cutout_lims = lims


    def _get_cutout_wcs(self):
        """
        Transform the cube WCS object into the cutout column WCS keywords.
        Adds the physical keywords for transformation back from cutout to location on FFI.
        Also performs nonliner -> linear approximation for the transformation matrix (TODO: not yet is doesn't!)
        This is a very TESS specific function.
        
        Returns
        -------
        response: dict
            Cutout wcs column header keywords as dictionary of 
            ``{<kwd format string>: [value, desc]} pairs.``
        """
        cube_wcs_header = self.cube_wcs.to_header(relax=True)

        cutout_wcs_dict = dict()

        ## Cutout array keywords ##
    
        cutout_wcs_dict["WCAX{}"] = [cube_wcs_header['WCSAXES'],"number of WCS axes"]
        # TODO: check for 2? this must be two

        cutout_wcs_dict["1CTYP{}"] = [cube_wcs_header["CTYPE1"],"right ascension coordinate type"]
        cutout_wcs_dict["2CTYP{}"] = [cube_wcs_header["CTYPE2"],"declination coordinate type"]
        
        orig_pix = self.cube_wcs.all_world2pix(self.center_coord.ra.deg, self.center_coord.dec.deg, 1)
        cutout_wcs_dict["1CRPX{}"] = [float(orig_pix[0]) - self.cutout_lims[0,0],
                                      "[pixel] reference pixel along image axis 1"]
        cutout_wcs_dict["2CRPX{}"] = [float(orig_pix[1]) - self.cutout_lims[0,0],
                                      "[pixel] reference pixel along image axis 2"]
    
        cutout_wcs_dict["1CRVAL{}"] = [self.center_coord.ra.deg, "[deg] right ascension at reference pixel"]
        cutout_wcs_dict["2CRVAL{}"] = [self.center_coord.dec.deg, "[deg] declination at reference pixel"]
    
        cunits = self.cube_wcs.wcs.cunit
        cutout_wcs_dict["1CUNI{}"] = [str(cunits[0]), "physical unit in column dimension"]
        cutout_wcs_dict["2CUNI{}"] = [str(cunits[1]), "physical unit in row dimension"]

        px_scales = wcs.utils.proj_plane_pixel_scales(self.cube_wcs)
        cutout_wcs_dict["1CDLT{}"] = [px_scales[0], "[deg] pixel scale in RA dimension"]
        cutout_wcs_dict["2CDLT{}"] = [px_scales[1], "[deg] pixel scale in DEC dimension"]

        # TODO: THIS IS FILLER, HAVE TO FIGURE OUT HOW TO DO THE TRANSFORMATION FOR REAL
        cutout_wcs_dict["11PC{}"] = [1,"linear transformation matrix element - unfilled"]
        cutout_wcs_dict["12PC{}"] = [1,"linear transformation matrix element - unfilled"]
        cutout_wcs_dict["21PC{}"] = [1,"linear transformation matrix element - unfilled"]
        cutout_wcs_dict["22PC{}"] = [1,"linear transformation matrix element - unfilled"]

        ## Physical keywords ##
    
        cutout_wcs_dict["WCSN{}P"] = ["PHYSICAL", "table column WCS name"]
        cutout_wcs_dict["WCAX{}P"] = [2,"table column physical WCS dimensions"]
    
        cutout_wcs_dict["1CTY{}P"] = ["RAWX","table column physical WCS axis 1 type, CCD col"]
        cutout_wcs_dict["2CTY{}P"] = ["RAWY","table column physical WCS axis 2 type, CCD row"]
    
        cutout_wcs_dict["1CUN{}P"] = ["PIXEL","table column physical WCS axis 1 unit"]
        cutout_wcs_dict["2CUN{}P"] = ["PIXEL","table column physical WCS axis 2 unit"]
    
        cutout_wcs_dict["1CRV{}P"] = [int(cube_wcs_header["CRPIX1"]) - self.cutout_lims[0,0],
                                      "table column physical WCS ax 1 ref value"]
        cutout_wcs_dict["2CRV{}P"] = [int(cube_wcs_header["CRPIX1"]) - self.cutout_lims[1,0],
                                      "table column physical WCS ax 2 ref value"]

        # TODO: can we calculate these? or are they fixed?
        cutout_wcs_dict["1CDL{}P"] = [1.0,"table column physical WCS a1 step"]    
        cutout_wcs_dict["2CDL{}P"] = [1.0,"table column physical WCS a2 step"]
    
        cutout_wcs_dict["1CRP{}P"] = [1,"table column physical WCS a1 reference"]
        cutout_wcs_dict["2CRP{}P"] = [1,"table column physical WCS a2 reference"]

        return cutout_wcs_dict
    

    def _get_cutout(self, transposed_cube, verbose=True):
        """
        Making a cutout from an image/uncertainty cube that has been transposed 
        to have time on the longest axis.
        
        Parameters
        ----------
        transposed_cube : `numpy.array`
            Transposed image/uncertainty array.
        verbose :  bool
            Optional. If true intermediate information is printed. 

        Returns
        -------
        response :  `numpy.array`, `numpy.array`, `numpy.array`
            The untransposed image cutout array,
            the untransposeduncertainty cutout array,
            and the aperture array (an array the size of a single cutout 
            that is 1 where there is image data and 0 where there isn't)
        """

        # These limits are not guarenteed to be within the image footprint
        xmin,xmax = self.cutout_lims[1]
        ymin,ymax = self.cutout_lims[0]

        # Get the image array limits
        xmax_cube,ymax_cube,_,_ = transposed_cube.shape

        # Adjust limits and figuring out the padding
        padding = np.zeros((3,2),dtype=int)
        if xmin < 0:
            padding[1,0] = -xmin
            xmin = 0
        if ymin < 0:
            padding[2,0] = -ymin
            ymin = 0
        if xmax > xmax_cube:
            padding[1,1] = xmax - xmax_cube
            xmax = xmax_cube
        if ymax > ymax_cube:
            padding[2,1] = ymax - ymax_cube
            ymax = ymax_cube       
        
        # Doing the cutout
        cutout = transposed_cube[xmin:xmax,ymin:ymax,:,:]
    
        img_cutout = cutout[:,:,:,0].transpose((2,0,1))
        uncert_cutout = cutout[:,:,:,1].transpose((2,0,1))
    
        # Making the aperture array
        aperture = np.ones((xmax-xmin, ymax-ymin))

        # Adding padding to the cutouts so that it's the expected size
        if padding.any(): # only do if we need to pad
            img_cutout = np.pad(img_cutout, padding, 'constant', constant_values=np.nan)
            uncert_cutout = np.pad(uncert_cutout, padding, 'constant', constant_values=np.nan)
            aperture = np.pad(aperture, padding[1:], 'constant', constant_values=0)

        if verbose:
            print("Image cutout cube shape: {}".format(img_cutout.shape))
            print("Uncertainty cutout cube shape: {}".format(uncert_cutout.shape))
    
        return img_cutout, uncert_cutout, aperture


    def _update_primary_header(self, primary_header):
        """
        Updates the primary header for the cutout target pixel file by filling in 
        the object ra and dec with the central cutout coordinates and filling in
        the rest of the TESS target pixel file keywords wil 0/empty strings
        as we do not have access to this information.
        This is a TESS-specific function.

        Parameters
        ----------
        primary_header : `~astropy.io.fits.Header`
            The primary header from the cube file that will be modified in place for the cutout.
        """

        # Adding cutout specific headers
        primary_header['RA_OBJ'] = (self.center_coord.ra.deg,'[deg] right ascension')
        primary_header['DEC_OBJ'] = (self.center_coord.dec.deg,'[deg] declination')

        primary_header['TIMEREF'] = ('SOLARSYSTEM', 'barycentric correction applied to times')        
        primary_header['TASSIGN'] = ('SPACECRAFT', 'where time is assigned')                         
        primary_header['TIMESYS'] = ('TDB', 'time system is Barycentric Dynamical Time (TDB)')
        primary_header['BJDREFI'] = (2457000, 'integer part of BTJD reference date')           
        primary_header['BJDREFF'] = (0.00000000, 'fraction of the day in BTJD reference date')    
        primary_header['TIMEUNIT'] = ('d', 'time unit for TIME, TSTART and TSTOP')

        telapse = primary_header.get("TSTOP",0) - primary_header.get("TSTART",0)
        primary_header['TELAPSE '] = (telapse, '[d] TSTOP - TSTART')
        
        # These are all the things in the TESS pipeline tpfs about the object that we can't fill
        primary_header['OBJECT'] = ("",'string version of target id ')
        primary_header['TCID'] = (0,'unique tess target identifier')
        primary_header['PXTABLE'] = (0,'pixel table id') 
        primary_header['PMRA'] = (0.0,'[mas/yr] RA proper motion') 
        primary_header['PMDEC'] = (0.0,'[mas/yr] Dec proper motion') 
        primary_header['PMTOTAL'] = (0.0,'[mas/yr] total proper motion') 
        primary_header['TESSMAG'] = (0.0,'[mag] TESS magnitude') 
        primary_header['TEFF'] = (0.0,'[K] Effective temperature') 
        primary_header['LOGG'] = (0.0,'[cm/s2] log10 surface gravity') 
        primary_header['MH'] =(0.0,'[log10([M/H])] metallicity') 
        primary_header['RADIUS'] = (0.0,'[solar radii] stellar radius')
        primary_header['TICVER'] = (0,'TICVER')

        

    def _add_column_wcs(self, table_header, wcs_dict):
        """
        Adds WCS information for the array columns to the cutout table header.

        Parameters
        ----------
        table_header : `~astropy.io.fits.Header`
            The table header for the cutout table that will be modified in place to include 
            WCS information.
        wcs_dict : dict
            Dictionary of wcs keyword/value pairs to be added to each array column in the 
            cutout table header.
        """

        wcs_col = False # says if column is one that requires wcs info
    
        for kwd in table_header:

            # Adding header descriptions for the table keywords
            if "TTYPE" in kwd:
                table_header.comments[kwd] = "column name"
            elif "TFORM" in kwd:
                table_header.comments[kwd] = "column format"
            elif "TUNIT" in kwd:
                table_header.comments[kwd] = "unit"
            elif "TDISP" in kwd:
                table_header.comments[kwd] = "display format"
            elif "TDIM" in kwd:
                table_header.comments[kwd] = "multi-dimensional array spec"
                wcs_col = True # if column holds 2D array need to add wcs info
            elif "TNULL" in kwd:
                table_header.comments[kwd] = "null value"

            # Adding wcs info if necessary
            if (kwd[:-1] == "TTYPE") and wcs_col:
                wcs_col = False # reset
                for wcs_key,(val,com) in wcs_dict.items():
                    table_header.insert(kwd,(wcs_key.format(int(kwd[-1])-1),val,com))

            # Removing keywords we don't want in the table header
            # TODO: figure out what these are (if any)

            
    def _add_aperture_wcs(self, aperture_header, cube_table_header):
        """
        Adds the full cube WCS information with adjustments for the cutout 
        to the aperture extension header in place.

        Parameters
        ----------
        aperture_header : `~astropy.io.fits.Header`
            The aperture extension header.  It will be modified in place.
        cube_table_header : `~astropy.io.fits.Header`
            The header for the table extension header from the cube file.
        """

        cube_wcs_header = self.cube_wcs.to_header(relax=True)

        # Adding the wcs keywords
        for kwd,val,cmt in cube_wcs_header.cards: 
            aperture_header.set(kwd,val,cube_table_header.get(kwd,cmt))
            # using table comment rather than the default ones if available

        # Adjusting the CRPIX/CRVAL values
        orig_pix = self.cube_wcs.all_world2pix(self.center_coord.ra.deg, self.center_coord.dec.deg, 1)
        aperture_header["CRPIX1"] = float(orig_pix[0]) - self.cutout_lims[0,0]
        aperture_header["CRPIX2"] = float(orig_pix[1]) - self.cutout_lims[1,0]
    
        aperture_header["CRVAL1"] = self.center_coord.ra.deg
        aperture_header["CRVAL2"] = self.center_coord.dec.deg

        # Adding the physical wcs keywords
        aperture_header.set("WCSNAMEP", "PHYSICAL","name of world coordinate system alternate P")
        aperture_header.set("WCSAXESP", 2, "number of WCS physical axes")
    
        aperture_header.set("CTYPE1P", "RAWX", "physical WCS axis 1 type CCD col")
        aperture_header.set("CUNIT1P", "PIXEL", "physical WCS axis 1 unit")
        aperture_header.set("CRPIX1P", 1, "reference CCD column")
        aperture_header.set("CRVAL1P", int(cube_wcs_header["CRPIX1"]) - self.cutout_lims[0,0], "value at reference CCD column")
        aperture_header.set("CDELT1P", 1.0, "physical WCS axis 1 step")
                
        aperture_header.set("CTYPE2P", "RAWY", "physical WCS axis 2 type CCD col")
        aperture_header.set("CUNIT2P", "PIXEL", "physical WCS axis 2 unit")
        aperture_header.set("CRPIX2P", 1, "reference CCD row")
        aperture_header.set("CRVAL2P", int(cube_wcs_header["CRPIX1"]) - self.cutout_lims[1,0], "value at reference CCD row")
        aperture_header.set("CDELT2P", 1.0, "physical WCS axis 2 step")

        

        

            
    def apply_header_inherit(self, hdu_list):
        """
        The INHERIT keyword indicated that keywords from the primary header should be duplicated in 
        the headers of all subsequent extensions.  This function performs this addition in place to 
        the given hdu list.
        
        Parameters
        ----------
        hdu_list : `~astropy.io.fits.HDUList`
            The hdu list to aaply the INHERIT keyword to.
        """
    
        primary_header = hdu_list[0].header

        reserved_kwds = ["COMMENT","SIMPLE","BITPIX", "EXTEND", "NEXTEND"]

        for hdu in hdu_list[1:]:
            if hdu.header.get("INHERIT", False):
                for kwd in primary_header:
                    if (kwd not in hdu.header) and (kwd not in reserved_kwds):
                        hdu.header[kwd] = (primary_header[kwd], primary_header.comments[kwd])
            

    def _build_tpf(self, cube_fits, img_cube, uncert_cube, cutout_wcs_dict, aperture, verbose=True):
        """
        Building the cutout target pixel file (TPF) and formatting it to match TESS pipeline TPFs.

        Paramters
        ---------
        cube_fits : `~astropy.io.fits.hdu.hdulist.HDUList`
            The cube hdu list.
        img_cube : `numpy.array`
            The untransposed image cutout array
        uncert_cube : `numpy.array`
            The untransposed uncertainty cutout array
        cutout_wcs_dict : dict
            Dictionary of wcs keyword/value pairs to be added to each array 
            column in the cutout table header.
        aperture : `numpy.array`
            The aperture array (an array the size of a single cutout 
            that is 1 where there is image data and 0 where there isn't)        
        verbose : bool
            Optional. If true intermediate information is printed. 

        Returns
        -------
        response :  `~astropy.io.fits.HDUList`
            Target pixel file HDU list
        """
        
        # The primary hdu is just the main header, which is the same
        # as the one on the cube file
        primary_hdu = cube_fits[0]
        self._update_primary_header(primary_hdu.header)

        cols = list()

        # Adding the cutouts
        tform = str(img_cube[0].size) + "E"
        dims = str(img_cube[0].shape)
        empty_arr = np.zeros(img_cube.shape)

        # Adding the Time relates columns
        cols.append(fits.Column(name='TIME', format='D', unit='BJD - 2457000, days', disp='D14.7',
                                array=(cube_fits[2].columns['TSTART'].array + cube_fits[2].columns['TSTOP'].array)/2))

        cols.append(fits.Column(name='TIMECORR', format='E', unit='d', disp='E14.7',
                                array=cube_fits[2].columns['BARYCORR'].array))

        # Adding CADENCENO as zeros b/c we don't have this info
        cols.append(fits.Column(name='CADENCENO', format='J', disp='I10', array=empty_arr[:,0,0]))

        # Adding counts (-1 b/c we don't have data)
        cols.append(fits.Column(name='RAW_CNTS', format=tform.replace('E','J'), unit='count', dim=dims, disp='I8',
                            array=empty_arr-1, null=-1))

        # Adding flux and flux_err (data we actually have!)
        cols.append(fits.Column(name='FLUX', format=tform, dim=dims, unit='e-/s', disp='E14.7', array=img_cube))
        cols.append(fits.Column(name='FLUX_ERR', format=tform, dim=dims, unit='e-/s', disp='E14.7', array=uncert_cube)) 
   
        # Adding the background info (zeros b.c we don't have this info)
        cols.append(fits.Column(name='FLUX_BKG', format=tform, dim=dims, unit='e-/s', disp='E14.7',array=empty_arr))
        cols.append(fits.Column(name='FLUX_BKG_ERR', format=tform, dim=dims, unit='e-/s', disp='E14.7',array=empty_arr))

        # Adding the quality flags
        cols.append(fits.Column(name='QUALITY', format='J', disp='B16.16', array=cube_fits[2].columns['DQUALITY'].array))

        # Adding the position correction info (zeros b.c we don't have this info)
        cols.append(fits.Column(name='POS_CORR1', format='E', unit='pixel', disp='E14.7',array=empty_arr[:,0,0]))
        cols.append(fits.Column(name='POS_CORR2', format='E', unit='pixel', disp='E14.7',array=empty_arr[:,0,0]))

        # Adding the FFI_FILE column (not in the pipeline tpfs)
        cols.append(fits.Column(name='FFI_FILE', format='38A', unit='pixel',array=cube_fits[2].columns['FFI_FILE'].array))
        
        # making the table HDU
        table_hdu = fits.BinTableHDU.from_columns(cols)
        table_hdu.header['EXTNAME'] = 'PIXELS'
        table_hdu.header['INHERIT'] = True
    
        # Adding the wcs keywords to the columns and removing from the header
        self._add_column_wcs(table_hdu.header, cutout_wcs_dict) 

        # Building the aperture HDU
        aperture_hdu = fits.ImageHDU(data=aperture)
        aperture_hdu.header['EXTNAME'] = 'APERTURE'
        self._add_aperture_wcs(aperture_hdu.header, cube_fits[2].header)
        aperture_hdu.header['INHERIT'] = True
    
        cutout_hdu_list = fits.HDUList([primary_hdu,table_hdu, aperture_hdu])
        
        self.apply_header_inherit(cutout_hdu_list)

        return cutout_hdu_list



    def cube_cut(self, cube_file, coordinates, cutout_size,
                 target_pixel_file=None, output_path=".", verbose=False):
        """
        Takes a cube file (as created by `~astrocut.CubeFactory`), and makes a cutout target pixel 
        file of the given size around the given coordinates. The target pixel file is formatted like
        a TESS pipeline target pixel file.

        Parameters
        ----------
        cube_file : str
            The cube file containing all the images to be cutout.  
            Must be in the format returned by ~astrocut.make_cube.
        coordinates : str or `astropy.coordinates.SkyCoord` object
            The position around which to cutout. 
            It may be specified as a string ("ra dec" in degrees) 
            or as the appropriate `~astropy.coordinates.SkyCoord` object.
        cutout_size : int, array-like, `~astropy.units.Quantity`
            The size of the cutout array. If ``cutout_size``
            is a scalar number or a scalar `~astropy.units.Quantity`,
            then a square cutout of ``cutout_size`` will be created.  If
            ``cutout_size`` has two elements, they should be in ``(ny, nx)``
            order.  Scalar numbers in ``cutout_size`` are assumed to be in
            units of pixels. `~astropy.units.Quantity` objects must be in pixel or
            angular units.
        target_pixel_file : str
            Optional. The name for the output target pixel file. 
            If no name is supplied, the file will be named: 
            ``<cube_file_base>_<ra>_<dec>_<cutout_size>_astrocut.fits``
        output_path : str
            Optional. The path where the output file is saved. 
            The current directory is default.
        verbose : bool
            Optional. If true intermediate information is printed. 

        Returns
        -------
        response: string or None
            If successfull, returns the path to the target pixel file, 
            if unsuccessful returns None.
        """

        if verbose:
            start_time = time()

        cube = fits.open(cube_file) 

        # Get the WCS and figure out which pixels are in the cutout
        wcs_ind = int(len(cube[2].data)/2) # using the middle file for wcs info
        self._get_cube_wcs(cube[2].header, cube[2].data[wcs_ind])

        if isinstance(coordinates, SkyCoord):
            self.center_coord = coordinates
        else:
            self.center_coord = SkyCoord(coordinates,unit='deg')

        if verbose:
            print("Cutout center coordinate: {},{}".format(self.center_coord.ra.deg,self.center_coord.dec.deg))


        # Making size into an array [ny, nx] 
        cutout_size = np.atleast_1d(cutout_size)
        if len(cutout_size) == 1:
            cutout_size = np.repeat(cutout_size, 2)

        if len(cutout_size) > 2:
            warnings.warn("Too many dimensions in cutout size, only the first two will be used.",
                          InputWarning)
       
        # Get cutout limits
        self._get_cutout_limits(cutout_size)

        if verbose:
            print("xmin,xmax: {}".format(self.cutout_lims[1]))
            print("ymin,ymax: {}".format(self.cutout_lims[0]))

        # Make the cutout
        img_cutout, uncert_cutout, aperture = self._get_cutout(cube[1].data, verbose=verbose)

        # Get cutout wcs info
        cutout_wcs_dict = self._get_cutout_wcs()
    
        # Build the TPF
        tpf_object = self._build_tpf(cube, img_cutout, uncert_cutout, cutout_wcs_dict, aperture)

        if verbose:
            write_time = time()

        if not target_pixel_file:
            _, flename = os.path.split(cube_file)
            target_pixel_file = output_path + "/"
            target_pixel_file += "{}_{}_{}_{}x{}_astrocut.fits".format(flename.rstrip('.fits').rstrip("-cube"),
                                                                       self.center_coord.ra.value,
                                                                       self.center_coord.dec.value,
                                                                       self.cutout_lims[0,1]-self.cutout_lims[0,0],
                                                                       self.cutout_lims[1,1]-self.cutout_lims[1,0])
        
        if verbose:
            print("Target pixel file:",target_pixel_file)
        
        # Write the TPF
        tpf_object.writeto(target_pixel_file, overwrite=True)

        # Close the cube file
        cube.close()

        if verbose:
            print("Write time: {:.2} sec".format(time()-write_time))
            print("Total time: {:.2} sec".format(time()-start_time))

        return target_pixel_file




# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module implements the functionality to create image cubes for the purposes of
creating cutout target pixel files.
"""

import numpy as np

from astropy.io import fits
from astropy.table import Table,Column

from time import time
from datetime import date
from os import path
from copy import deepcopy


class CubeFactory():
    """
    Class for creating image cubes.
    
    This class emcompasses all of the cube making functionality.  
    In the current version this means creating image cubes fits files from TESS full frame image sets.
    Future versions will include more generalized cubing functionality.
    """
                
    def _make_primary_header(self, ffi_main_header, ffi_img_header, sector=-1):
        """
        Given the primary and image headers from an input FFI, and the sector number, 
        build the cube's primary header.
        This is a TESS specific function

        Parameters
        ----------
        ffi_main_header : `~astropy.io.fits.Header`
            The primary header from a TESS FFI fits file.
        ffi_img_header : `~astropy.io.fits.Header`
            The seconday (image) header from a TESS FFI fits file.
        sector : int
            TESS mission sector number

        Returns
        -------
        response :  `~astropy.io.fits.Header`
            Primary header for the image cube fits file.
        """
    
        header = deepcopy(ffi_main_header)
        header.remove('CHECKSUM', ignore_missing=True)

        header['ORIGIN'] = 'STScI/MAST'
        header['DATE'] = str(date.today())
        header['SECTOR'] = (sector, "Observing sector")
        header['CAMERA'] = (ffi_img_header['CAMERA'], ffi_img_header.comments['CAMERA'])
        header['CCD'] = (ffi_img_header['CCD'], ffi_img_header.comments['CCD'])

        

        return header
    
    def make_cube(self, file_list, cube_file="img-cube.fits", sector=None, verbose=True):
        """
        Turns a list of fits image files into one large data-cube.
        Input images must all have the same footprint and resolution.
        The resulting datacube is transposed for quicker cutouts.
        This function can take some time to run and requires enough 
        memory to hold the entire cube in memory. 
        (For full TESS sectors this is about 40 GB)

        Parameters
        ----------
        file_list : array
            The list of fits image files to cube.  
            Assumed to have the format of a TESS FFI:
            - A primary HDU consisting only of a primary header
            - An image HDU containing the image
            - A second image HDU containing the uncertainty image
        cube_file : string
            Optional.  The filename/path to save the output cube in. 
        sector : int
            Optional.  TESS sector to add as header keyword (not present in FFI files).
        verbose : bool
            Optional. If true intermediate information is printed. 

        Returns
        -------
        response: string or None
            If successful, returns the path to the cube fits file, 
            if unsuccessful returns None.
        """

        if verbose:
            startTime = time()

        # Getting the time sorted indices for the files
        file_list = np.array(file_list)
        start_times = np.zeros(len(file_list))
    
        for i,ffi in enumerate(file_list):
            start_times[i] = fits.getval(ffi, 'TSTART', 1) # TODO: optionally pass this in?
            
        sorted_indices = np.argsort(start_times)      
    
        # these will be the arrays and headers
        img_cube = None
        img_info_table = None
        primary_header = None
        secondary_header = None

        # Loop through files
        for i,fle in enumerate(file_list[sorted_indices]):
        
            ffi_data = fits.open(fle)

            # if the arrays/headers aren't initialized do it now
            if img_cube is None:

                # We use the primary header from the first file as the cube primary header
                # and will add in information about the time of the final observation at the end
                primary_header = self._make_primary_header(ffi_data[0].header, ffi_data[1].header, sector=sector)

                # The image specific header information will be saved in a table in the second extension
                secondary_header = ffi_data[1].header
            
                ffi_img = ffi_data[1].data

                # set up the cube array
                img_cube = np.full((ffi_img.shape[0],ffi_img.shape[1],len(file_list),2),
                                   np.nan, dtype=np.float32)

                # set up the image info table
                cols = []
                for kwd,val,cmt in secondary_header.cards: 
                    if type(val) == str:
                        tpe = "S"+str(len(val))
                    elif type(val) == int:
                        tpe = np.int32
                    else:
                        tpe = np.float32
                    
                    cols.append(Column(name=kwd,dtype=tpe,length=len(file_list), meta={"comment":cmt}))
                    
                cols.append(Column(name="FFI_FILE",dtype="S" + str(len(path.basename(fle))), length=len(file_list)))
            
                img_info_table = Table(cols)

            # add the image and info to the arrays
            img_cube[:,:,i,0] = ffi_data[1].data
            img_cube[:,:,i,1] = ffi_data[2].data
            for kwd in img_info_table.columns:
                if kwd == "FFI_FILE":
                    img_info_table[kwd][i] = path.basename(fle)
                else:
                    img_info_table[kwd][i] = ffi_data[1].header[kwd]

            #if fle == file_list[-1]:
            if i == (len(file_list) - 1):
                primary_header['DATE-END'] = ffi_data[0].header['DATE-END']
                primary_header['TSTOP'] = ffi_data[0].header.get('TSTOP', 0)
                
            # close fits file
            ffi_data.close()

            if verbose:
                print("Completed file {}".format(i))

        # put it all in a fits file 
        primary_hdu = fits.PrimaryHDU(header=primary_header)
        cube_hdu = fits.ImageHDU(data=img_cube)

        # make table hdu with the img info array
        cols = []
        for kwd in img_info_table.columns:
            if img_info_table[kwd].dtype == np.float32:
                tpe = 'D'
            elif img_info_table[kwd].dtype == np.int32:
                tpe = 'J'
            else:
                tpe = str(img_info_table[kwd].dtype).replace("S","A").strip("|")
        
            cols.append(fits.Column(name=kwd, format=tpe, array=img_info_table[kwd]))
        
        col_def = fits.ColDefs(cols)
        table_hdu = fits.BinTableHDU.from_columns(col_def)

        # Adding the comments to the header
        for kwd in img_info_table.columns:
            if kwd in ['XTENSION', 'BITPIX', 'NAXIS','NAXIS1','NAXIS2','PCOUNT','GCOUNT','TFIELDS']:
                continue # skipping the keyword already in use
            table_hdu.header[kwd] = img_info_table[kwd].meta.get("comment","")

        hdu_list = fits.HDUList([primary_hdu,cube_hdu,table_hdu])

        if verbose:
            writeTime = time()
    
        hdu_list.writeto(cube_file, overwrite=True) 

        if verbose:
            endTime = time()
            print("Total time elapsed: {:.2f} sec".format(endTime - startTime))
            print("File write time: {:.2f} sec".format(endTime - writeTime))

        return cube_file



import numpy as np

from astropy.io import fits
from astropy.table import Table

from .utils_for_test import create_test_ffis
from ..make_cube import CubeFactory



def test_make_cube(tmpdir):
    """
    Testing the make cube functionality by making a bunch of test FFIs, 
    making the cube, and checking the results.
    """

    cube_maker = CubeFactory()
    
    img_sz = 10
    num_im = 100
    
    ffi_files = create_test_ffis(img_sz, num_im)
    cube_file = cube_maker.make_cube(ffi_files, "make_cube-test-cube", verbose=False)

    hdu = fits.open(cube_file)
    cube = hdu[1].data
    
    # expected values for cube
    ecube = np.zeros((img_sz,img_sz,num_im,2))
    plane = np.arange(img_sz*img_sz,dtype=np.float32).reshape((img_sz,img_sz))
    assert cube.shape == ecube.shape, "Mismatch between cube shape and expected shape"

    for i in range(num_im):
        ecube[:,:,i,0] = -plane
        ecube[:,:,i,1] = plane
        plane += img_sz*img_sz

    assert np.alltrue(cube == ecube), "Cube values do not match expected values"
        
    tab = Table(hdu[2].data)
    assert np.alltrue(tab['TSTART'] == np.arange(num_im)), "TSTART mismatch in table"
    assert np.alltrue(tab['TSTOP'] == np.arange(num_im)+1), "TSTOP mismatch in table"
    assert np.alltrue(tab['FFI_FILE'] == np.array(ffi_files)), "FFI_FILE mismatch in table"

    hdu.close()
    



import os
import numpy as np

from astropy.io import fits
from astropy.table import Table


def add_keywords(hdu, extname, time_increment, primary=False):
    """
    Add a bunch of required keywords (mostly fake values)
    """
    
    hdu.header['extname'] = 'CAMERA.CCD 1.1 cal'
    hdu.header['camera'] = 1
    hdu.header['ccd'] = 1
    hdu.header['tstart'] = float(time_increment)
    hdu.header['tstop'] = float(time_increment+1)
    hdu.header['date-obs'] = '2019-05-11T00:08:26.816Z'
    hdu.header['date-end'] = '2019-05-11T00:38:26.816Z'
    hdu.header['barycorr'] = 5.0085597E-03
    hdu.header['dquality'] = 0

    if not primary:
        # WCS keywords just copied from example
        hdu.header['RADESYS'] = 'ICRS    '
        hdu.header['EQUINOX'] =               2000.0
        hdu.header['WCSAXES'] =                    2
        hdu.header['CTYPE1']  = 'RA---TAN-SIP'
        hdu.header['CTYPE2']  = 'DEC--TAN-SIP'
        hdu.header['CRVAL1']  = 250.3497414839765200
        hdu.header['CRVAL2']  =   2.2809255996090630
        hdu.header['CRPIX1']  =               1045.0
        hdu.header['CRPIX2']  =               1001.0
        hdu.header['CD1_1']   =   -0.005564478186178
        hdu.header['CD1_2']   =   -0.001042099258152
        hdu.header['CD2_1']   =    0.001181441465850
        hdu.header['CD2_2']   =   -0.005590816683583
        hdu.header['A_ORDER'] =                    2
        hdu.header['B_ORDER'] =                    2
        hdu.header['A_2_0']   =   2.024511892340E-05
        hdu.header['A_0_2']   =   3.317603337918E-06
        hdu.header['A_1_1']   =  1.73456334971071E-5
        hdu.header['B_2_0']   =   3.331330003472E-06
        hdu.header['B_0_2']   = 2.042474824825892E-5
        hdu.header['B_1_1']   = 1.714767108041439E-5
        hdu.header['AP_ORDER']=                    2
        hdu.header['BP_ORDER']=                    2
        hdu.header['AP_1_0']  = 9.047002963896363E-4
        hdu.header['AP_0_1']  = 6.276607155847164E-4
        hdu.header['AP_2_0']  =  -2.023482905861E-05
        hdu.header['AP_0_2']  =  -3.332285841011E-06
        hdu.header['AP_1_1']  =  -1.731636633824E-05
        hdu.header['BP_1_0']  = 6.279608820532116E-4
        hdu.header['BP_0_1']  = 9.112228860848081E-4
        hdu.header['BP_2_0']  =  -3.343918167224E-06
        hdu.header['BP_0_2']  =  -2.041598249021E-05
        hdu.header['BP_1_1']  =  -1.711876336719E-05
        hdu.header['A_DMAX']  =    44.72893589844534
        hdu.header['B_DMAX']  =    44.62692873032506


def create_test_ffis(img_size, num_images, basename='make_cube-test{:04d}.fits'):
    """
    Create test fits files

    Write negative values for data array and positive values for error array,
    with unique values for all the pixels. The header keywords are just
    populated to make things run but are not tested.
    """

    img = np.arange(img_size*img_size, dtype=np.float32).reshape((img_size,img_size))
    file_list = []
    
    for i in range(num_images):
        
        file_list.append(basename.format(i))

        primary_hdu = fits.PrimaryHDU()
        add_keywords(primary_hdu, "PRIMARY", i, primary=True)
        
        
        hdu = fits.ImageHDU(-img)
        add_keywords(hdu, 'CAMERA.CCD 1.1 cal', i)
        
        ehdu = fits.ImageHDU(img)
        add_keywords(ehdu, 'CAMERA.CCD 1.1 uncert', i)
        
        hdulist = fits.HDUList([primary_hdu, hdu, ehdu])
        hdulist.writeto(file_list[-1], overwrite=True, checksum=True)
        
        img = img + img_size*img_size

    return file_list



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Custom exceptions used in the astrocut classes
"""

from astropy.utils.exceptions import AstropyWarning


class InvalidQueryError(Exception):
    """
    Errors related to invalid queries.
    """
pass


class InputWarning(AstropyWarning):
    """
    Warning to be issued when use input is incorrect in
    some way but doesn't prevent the function from running.
    """
pass

class TypeWarning(AstropyWarning):
    """
    Warnings to do with data types.
    """
pass
